@@ -1,44 +1,28 @@
 /*
  * pg_dbms_stats.c
  *
- * Copyright (c) 2009-2018, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
- * Portions Copyright (c) 1996-2013, PostgreSQL Global Development Group
+ * Copyright (c) 2009-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  */
 #include "postgres.h"
 
-#include "access/sysattr.h"
 #include "access/transam.h"
-#include "catalog/pg_index.h"
 #include "catalog/pg_statistic.h"
 #include "catalog/pg_type.h"
-#include "catalog/namespace.h"
-#include "catalog/pg_authid.h"
 #include "commands/trigger.h"
 #include "executor/spi.h"
 #include "funcapi.h"
 #include "optimizer/plancat.h"
-#include "optimizer/planner.h"
-#include "parser/parse_oper.h"
-#include "parser/parsetree.h"
 #include "storage/bufmgr.h"
-#include "utils/acl.h"
 #include "utils/builtins.h"
-#include "utils/elog.h"
-#include "utils/fmgroids.h"
 #include "utils/guc.h"
 #include "utils/inval.h"
 #include "utils/lsyscache.h"
 #include "utils/selfuncs.h"
 #include "utils/syscache.h"
-#include "miscadmin.h"
+#if PG_VERSION_NUM >= 90200
 #include "utils/rel.h"
-#if PG_VERSION_NUM >= 90300
-#include "access/htup_details.h"
-#include "utils/catcache.h"
-#endif
-#if PG_VERSION_NUM >= 100000
-#include <math.h>
 #endif
 
 #include "pg_dbms_stats.h"
@@ -51,83 +35,43 @@ PG_MODULE_MAGIC;
 
 #define MAX_REL_CACHE		50		/* expected max # of rel stats entries */
 
-/* acl_ok must be set after the fix for CVE-2017-7484 */
-#if PG_VERSION_NUM >= 100000 || \
-	(PG_VERSION_NUM >=  90603 && PG_VERSION_NUM < 100000) ||	\
-	(PG_VERSION_NUM >=  90507 && PG_VERSION_NUM <  90600) ||	\
-	(PG_VERSION_NUM >=  90412 && PG_VERSION_NUM <  90500) ||	\
-	(PG_VERSION_NUM >=  90317 && PG_VERSION_NUM <  90400) ||	\
-	(PG_VERSION_NUM >=  90221 && PG_VERSION_NUM <  90300)
-#define PGDS_HAVE_ACL_OK
-/*
- * acl_ok of the returning VariableStatData must be set if set_acl_okk is
- * true. The code is compiled only if the compile target PG version is match
- * the above conditions. Conversely, acl_ok is added to the end of
- * VariableStatData so we can safely omit setting it when the PG version
- * pg_dbms_stats is loaded onto is out of the conditions.
- */
-static	bool set_acl_ok = false;
-#endif
-
-/* FormData_pg_attribute is modifed at PG11 */
-#if PG_VERSION_NUM >= 110000
-#define get_attrs(pgatt_attrs) (&(pgatt_attrs))
-#else
-#define get_attrs(pgatt_attrs) (pgatt_attrs)
-#endif
-
-/* Relation statistics cache entry */
+/* represents an element of relation statistics which is cached in hash table */
 typedef struct StatsRelationEntry
 {
 	Oid					relid;		/* hash key must be at the head */
+
 	bool				valid;		/* T if the entry has valid stats */
-	bool				invalidated; /* T if this relation has been
-									  * invalidated */
+
 	BlockNumber			relpages;	/* # of pages as of last ANALYZE */
 	double				reltuples;	/* # of tuples as of last ANALYZE */
 	BlockNumber			relallvisible;	/* # of all-visible pages as of last
 										 * ANALYZE */
 	BlockNumber			curpages;	/* # of pages as of lock/restore */
-	List			   *col_stats;	/* list of StatsColumnEntry, each element
-									   of which is pg_statistic record of this
-									   relation. */
-} StatsRelationEntry;
 
-/*
- * Column statistics cache entry. This is for list item for
- * StatsRelationEntry.col_stats.
- */
-typedef struct StatsColumnEntry
-{
-  bool		negative;
-  int32		attnum;
-  bool		inh;
-  HeapTuple tuple;
-} StatsColumnEntry;
+	List			   *col_stats;	/* list of HeapTuple, each element is
+									   pg_statistic record of this relation. */
+} StatsRelationEntry;
 
 /* Saved hook functions */
 get_relation_info_hook_type		prev_get_relation_info = NULL;
 get_attavgwidth_hook_type		prev_get_attavgwidth = NULL;
 get_relation_stats_hook_type	prev_get_relation_stats = NULL;
 get_index_stats_hook_type		prev_get_index_stats = NULL;
-planner_hook_type				prev_planner_hook = NULL;
-
-/* namings */
-#define NSPNAME "dbms_stats"
-#define RELSTAT_TBLNAME "relation_stats_locked"
-#define COLSTAT_TBLNAME "column_stats_locked"
 
 /* rows_query(oid) RETURNS int4, float4, int4 */
 static const char  *rows_query =
-	"SELECT relpages, reltuples, curpages, relallvisible"
-	"  FROM " NSPNAME "." RELSTAT_TBLNAME
+	"SELECT relpages, reltuples, curpages"
+#if PG_VERSION_NUM >= 90200
+	", relallvisible"
+#endif
+	"  FROM dbms_stats.relation_stats_locked"
 	" WHERE relid = $1";
 static SPIPlanPtr	rows_plan = NULL;
 
 /* tuple_query(oid, int2, bool) RETURNS pg_statistic */
 static const char  *tuple_query =
 	"SELECT * "
-	"  FROM " NSPNAME "." COLSTAT_TBLNAME
+	"  FROM dbms_stats.column_stats_locked "
 	" WHERE starelid = $1 "
 	"   AND staattnum = $2 "
 	"   AND stainherit = $3";
@@ -141,16 +85,8 @@ static int			nested_level = 0;
 
 /*
  * The relation_stats_effective statistic cache is stored in hash table.
- * rel_invalidated is set true if the hash has invalidated entries.
  */
 static HTAB	   *rel_stats;
-static bool		rel_invalidated = false;
-
-/*
- * The owner of pg_dbms_stats statistic tables.
- */
-static Oid	stats_table_owner = InvalidOid;
-static char *stats_table_owner_name = "";
 
 #define get_pg_statistic(tuple)	((Form_pg_statistic) GETSTRUCT(tuple))
 
@@ -159,18 +95,12 @@ PG_FUNCTION_INFO_V1(dbms_stats_invalidate_relation_cache);
 PG_FUNCTION_INFO_V1(dbms_stats_invalidate_column_cache);
 PG_FUNCTION_INFO_V1(dbms_stats_is_system_schema);
 PG_FUNCTION_INFO_V1(dbms_stats_is_system_catalog);
-PG_FUNCTION_INFO_V1(dbms_stats_anyary_anyary);
-PG_FUNCTION_INFO_V1(dbms_stats_type_is_analyzable);
-PG_FUNCTION_INFO_V1(dbms_stats_anyarray_basetype);
 
 extern Datum dbms_stats_merge(PG_FUNCTION_ARGS);
 extern Datum dbms_stats_invalidate_relation_cache(PG_FUNCTION_ARGS);
 extern Datum dbms_stats_invalidate_column_cache(PG_FUNCTION_ARGS);
 extern Datum dbms_stats_is_system_schema(PG_FUNCTION_ARGS);
 extern Datum dbms_stats_is_system_catalog(PG_FUNCTION_ARGS);
-extern Datum dbms_stats_anyary_anyary(PG_FUNCTION_ARGS);
-extern Datum dbms_stats_type_is_analyzable(PG_FUNCTION_ARGS);
-extern Datum dbms_stats_anyarray_basetype(PG_FUNCTION_ARGS);
 
 static HeapTuple dbms_stats_merge_internal(HeapTuple lhs, HeapTuple rhs,
 	TupleDesc tupledesc);
@@ -182,7 +112,6 @@ static void dbms_stats_invalidate_cache_internal(Oid relid, bool sta_col);
 void	_PG_init(void);
 void	_PG_fini(void);
 
-/* hook functions */
 static void dbms_stats_get_relation_info(PlannerInfo *root, Oid relid,
 	bool inhparent, RelOptInfo *rel);
 static int32 dbms_stats_get_attavgwidth(Oid relid, AttrNumber attnum);
@@ -190,26 +119,19 @@ static bool dbms_stats_get_relation_stats(PlannerInfo *root, RangeTblEntry *rte,
 	AttrNumber attnum, VariableStatData *vardata);
 static bool dbms_stats_get_index_stats(PlannerInfo *root, Oid indexOid,
 	AttrNumber indexattnum, VariableStatData *vardata);
-static PlannedStmt *dbms_stats_planner(Query *parse, int cursorOptions,
-									   ParamListInfo boundParams);
 
-/* internal functions */
 static void get_merged_relation_stats(Oid relid, BlockNumber *pages,
 	double *tuples, double *allvisfrac, bool estimate);
 static int32 get_merged_avgwidth(Oid relid, AttrNumber attnum);
 static HeapTuple get_merged_column_stats(Oid relid, AttrNumber attnum,
 	bool inh);
-static HeapTuple column_cache_search(Oid relid, AttrNumber attnum,
-									 bool inh, bool*negative);
-static HeapTuple column_cache_enter(Oid relid, int32 attnum, bool inh,
-									HeapTuple tuple);
+static HeapTuple column_cache_search(Oid relid, AttrNumber attnum, bool inh);
+static HeapTuple column_cache_enter(HeapTuple tuple);
 static bool execute_plan(SPIPlanPtr *plan, const char *query, Oid relid,
 	const AttrNumber *attnum, bool inh);
-static void statscache_rel_callback(Datum arg, Oid relid);
-static void cleanup_invalidated_cache(void);
+static void StatsCacheRelCallback(Datum arg, Oid relid);
 static void init_rel_stats(void);
 static void init_rel_stats_entry(StatsRelationEntry *entry, Oid relid);
-
 /* copied from PG core source tree */
 static void dbms_stats_estimate_rel_size(Relation rel, int32 *attr_widths,
 				  BlockNumber *pages, double *tuples, double *allvisfrac,
@@ -243,25 +165,6 @@ _PG_init(void)
 	}
 #endif
 
-#ifdef PGDS_HAVE_ACL_OK
-	{
-		/*
-		 * Check the PG version this module loaded onto. This aid is required
-		 * for binary backward compatibility within a major PG version.
-		 */
-		int major_version = PG_VERSION_NUM / 100;
-		int minor_version = PG_VERSION_NUM % 100;
-
-		if (major_version >= 1000 ||
-			(major_version == 906 && minor_version >= 3) ||
-			(major_version == 905 && minor_version >= 7) ||
-			(major_version == 904 && minor_version >= 12) ||
-			(major_version == 903 && minor_version >= 17) ||
-			(major_version == 902 && minor_version >= 21))
-			set_acl_ok = true;
-	}
-#endif
-
 	/* Define custom GUC variables. */
 	DefineCustomBoolVariable("pg_dbms_stats.use_locked_stats",
 							 "Enable user defined statistics.",
@@ -285,14 +188,12 @@ _PG_init(void)
 	get_relation_stats_hook = dbms_stats_get_relation_stats;
 	prev_get_index_stats = get_index_stats_hook;
 	get_index_stats_hook = dbms_stats_get_index_stats;
-	prev_planner_hook = planner_hook;
-	planner_hook = dbms_stats_planner;
 
 	/* Initialize hash table for statistics caching. */
 	init_rel_stats();
 
 	/* Also set up a callback for relcache SI invalidations */
-	CacheRegisterRelcacheCallback(statscache_rel_callback, (Datum) 0);
+	CacheRegisterRelcacheCallback(StatsCacheRelCallback, (Datum) 0);
 }
 
 /*
@@ -306,108 +207,10 @@ _PG_fini(void)
 	get_attavgwidth_hook = prev_get_attavgwidth;
 	get_relation_stats_hook = prev_get_relation_stats;
 	get_index_stats_hook = prev_get_index_stats;
-	planner_hook = prev_planner_hook;
 
 	/* A function to unregister callback for relcache is NOT provided. */
 }
 
-/*
- * Function to convert from any array from dbms_stats.anyarray.
- */
-Datum
-dbms_stats_anyary_anyary(PG_FUNCTION_ARGS)
-{
-  ArrayType *arr = PG_GETARG_ARRAYTYPE_P(0);
-  if (ARR_NDIM(arr) != 1)
-	  elog(ERROR, "array must be one-dimentional.");
-
-  PG_RETURN_ARRAYTYPE_P(arr);
-}
-
-/*
- * Function to check if the type can have statistics.
- */
-Datum
-dbms_stats_type_is_analyzable(PG_FUNCTION_ARGS)
-{
-	Oid typid = PG_GETARG_OID(0);
-	Oid	eqopr;
-
-	if (!OidIsValid(typid))
-		PG_RETURN_BOOL(false);
-
-	get_sort_group_operators(typid, false, false, false,
-							 NULL, &eqopr, NULL,
-							 NULL);
-	PG_RETURN_BOOL(OidIsValid(eqopr));
-}
-
-/*
- * Function to get base type of the value of the type dbms_stats.anyarray.
- */
-Datum
-dbms_stats_anyarray_basetype(PG_FUNCTION_ARGS)
-{
-	ArrayType  *arr = PG_GETARG_ARRAYTYPE_P(0);
-	Oid			elemtype = arr->elemtype;
-	HeapTuple	tp;
-	Form_pg_type typtup;
-	Name		result;
-
-	if (!OidIsValid(elemtype))
-		elog(ERROR, "invalid base type oid: %u", elemtype);
-
-	tp = SearchSysCache1(TYPEOID, ObjectIdGetDatum(elemtype));
-	if (!HeapTupleIsValid(tp))  /* I trust you. */
-		elog(ERROR, "invalid base type oid: %u", elemtype);
-
-	typtup = (Form_pg_type) GETSTRUCT(tp);
-	result = (Name) palloc0(NAMEDATALEN);
-	StrNCpy(NameStr(*result), NameStr(typtup->typname), NAMEDATALEN);
-
-	ReleaseSysCache(tp);
-	PG_RETURN_NAME(result);
-}
-
-/*
- * Find and store the owner of the dummy statistics table.
- *
- * We will access statistics tables using this owner
- */
-static Oid
-get_stats_table_owner(void)
-{
-	HeapTuple tp;
-
-	if (!OidIsValid(stats_table_owner))
-	{
-		tp = SearchSysCache2(RELNAMENSP,
-					 PointerGetDatum(RELSTAT_TBLNAME),
-					 ObjectIdGetDatum(get_namespace_oid(NSPNAME, false)));
-		if (!HeapTupleIsValid(tp))
-			elog(ERROR, "table \"%s.%s\" not found in pg_class",
-				 NSPNAME, RELSTAT_TBLNAME);
-		stats_table_owner =	((Form_pg_class) GETSTRUCT(tp))->relowner;
-		if (!OidIsValid(stats_table_owner))
-			elog(ERROR, "owner uid of table \"%s.%s\" is invalid",
-				 NSPNAME, RELSTAT_TBLNAME);
-		ReleaseSysCache(tp);
-
-		tp = SearchSysCache1(AUTHOID, ObjectIdGetDatum(stats_table_owner));
-		if (!HeapTupleIsValid(tp))
-		{
-			elog(ERROR,
-				 "role id %u for the owner of the relation \"%s.%s\"is invalid",
-				 stats_table_owner, NSPNAME, RELSTAT_TBLNAME);
-		}
-		/* This will be done once for the session, so not pstrdup. */
-		stats_table_owner_name =
-			strdup(NameStr(((Form_pg_authid) GETSTRUCT(tp))->rolname));
-		ReleaseSysCache(tp);
-	}
-	return stats_table_owner;
-}
-
 /*
  * Store heap tuple header into given heap tuple.
  */
@@ -466,9 +269,6 @@ dbms_stats_merge(PG_FUNCTION_ARGS)
  * dbms_stats_merge_internal
  *   merge the dummy statistic (lhs) and the true statistic (rhs), on the basis
  *   of given TupleDesc.
- *
- *   this function doesn't become an error level of ERROR to meet that the 
- *   result of the SQL is not affected by the query plan.
  */
 static HeapTuple
 dbms_stats_merge_internal(HeapTuple lhs, HeapTuple rhs, TupleDesc tupdesc)
@@ -477,8 +277,6 @@ dbms_stats_merge_internal(HeapTuple lhs, HeapTuple rhs, TupleDesc tupdesc)
 	bool			nulls[Natts_pg_statistic];
 	int				i;
 	Oid				atttype = InvalidOid;
-	Oid				relid;
-	AttrNumber		attnum;
 
 	/* fast path for both-sides are null */
 	if ((lhs == NULL || lhs->t_data == NULL) &&
@@ -521,7 +319,7 @@ dbms_stats_merge_internal(HeapTuple lhs, HeapTuple rhs, TupleDesc tupdesc)
 						(errmsg("pg_dbms_stats: bad statistics"),
 						 errdetail("column \"%s\" should not be null",
 							get_attname(StatisticRelationId,
-									get_attrs(tupdesc->attrs[i])->attnum))));
+										tupdesc->attrs[i]->attnum))));
 					return NULL;	/* should not be null */
 				}
 			}
@@ -548,7 +346,7 @@ dbms_stats_merge_internal(HeapTuple lhs, HeapTuple rhs, TupleDesc tupdesc)
 							(errmsg("pg_dbms_stats: bad statistics"),
 							 errdetail("column \"%s\" should not be null",
 								get_attname(StatisticRelationId,
-									get_attrs(tupdesc->attrs[i])->attnum))));
+											tupdesc->attrs[i]->attnum))));
 						return NULL;	/* should not be null */
 					}
 				}
@@ -558,44 +356,27 @@ dbms_stats_merge_internal(HeapTuple lhs, HeapTuple rhs, TupleDesc tupdesc)
 		}
 	}
 
-	/*
-	 * Verify types to work around for ALTER COLUMN TYPE.
-	 *
-	 * Note: We don't need to retrieve atttype when the attribute doesn't have
-	 * neither Most-Common-Value nor Histogram, but we retrieve it always
-	 * because it's not usual.
-	 */
-	relid = DatumGetObjectId(values[0]);
-	attnum = DatumGetInt16(values[1]);
-	atttype = get_atttype(relid, attnum);
-	if (atttype == InvalidOid)
-	{
-		ereport(WARNING,
-		(errmsg("pg_dbms_stats: no-longer-existent column"),
-		 errdetail("relid \"%d\" or its column whose attnum is \"%d\" might be deleted",
-				relid, attnum),
-			 errhint("dbms_stats.clean_up_stats() would fix this.")));
-		return NULL;
-	}
+	/* verify types to work around for ALTER COLUMN TYPE */
 	for (i = 0; i < STATISTIC_NUM_SLOTS; i++)
 	{
 		if ((i + 1 == STATISTIC_KIND_MCV ||
 			 i + 1 == STATISTIC_KIND_HISTOGRAM) &&
 			!nulls[Anum_pg_statistic_stavalues1 + i - 1])
 		{
+			Oid			relid = DatumGetObjectId(values[0]);
+			AttrNumber	attnum = DatumGetInt16(values[1]);
 			ArrayType  *arr;
 
+			if (atttype == InvalidOid)
+				atttype = get_atttype(relid, attnum);
+
 			arr = DatumGetArrayTypeP(
 					values[Anum_pg_statistic_stavalues1 + i - 1]);
 			if (arr == NULL || arr->elemtype != atttype)
 			{
-				const char	   *attname = get_attname(relid, attnum);
+				char *attname = get_attname(relid,
+									tupdesc->attrs[attnum - 1]->attnum);
 
-				/*
-				 * relid and attnum must be valid here because valid atttype
-			 	 * has been gotten already.
-				 */
-				Assert(attname);
 				ereport(ELEVEL_BADSTATS,
 					(errmsg("pg_dbms_stats: bad column type"),
 					 errdetail("type of column \"%s\" has been changed",
@@ -779,7 +560,7 @@ dbms_stats_is_system_schema_internal(char *schema_name)
 	if (strcmp(schema_name, "pg_catalog") == 0 ||
 		strcmp(schema_name, "pg_toast") == 0 ||
 		strcmp(schema_name, "information_schema") == 0 ||
-		strcmp(schema_name, NSPNAME) == 0)
+		strcmp(schema_name, "dbms_stats") == 0)
 		return true;
 
 	return false;
@@ -873,8 +654,15 @@ dbms_stats_get_relation_info(PlannerInfo *root,
 	 * calculation.
 	 */
 	if (!inhparent)
+	{
+#if PG_VERSION_NUM >= 90200
 		get_merged_relation_stats(relid, &rel->pages, &rel->tuples,
 								  &rel->allvisfrac, true);
+#else
+		get_merged_relation_stats(relid, &rel->pages, &rel->tuples,
+								  &allvisfrac, true);
+#endif
+	}
 	else
 		return;
 
@@ -945,25 +733,10 @@ dbms_stats_get_relation_stats(PlannerInfo *root,
 		HeapTuple	tuple;
 
 		tuple = get_merged_column_stats(rte->relid, attnum, rte->inh);
-		vardata->statsTuple = tuple;
-		if (HeapTupleIsValid(tuple))
+		if (tuple != NULL)
 		{
+			vardata->statsTuple = tuple;
 			vardata->freefunc = FreeHeapTuple;
-
-#ifdef PGDS_HAVE_ACL_OK
-			/*
-			 * set acl_ok if required. See the definition of set_acl_ok for
-			 * details.
-			 */
-			if (set_acl_ok)
-			 {
-				 vardata->acl_ok =
-					 (pg_class_aclcheck(rte->relid, GetUserId(),
-										ACL_SELECT) == ACLCHECK_OK) ||
-					 (pg_attribute_aclcheck(rte->relid, attnum, GetUserId(),
-											ACL_SELECT) == ACLCHECK_OK);
-			 }
-#endif
 			return true;
 		}
 	}
@@ -986,103 +759,25 @@ dbms_stats_get_index_stats(PlannerInfo *root,
 						   AttrNumber indexattnum,
 						   VariableStatData *vardata)
 {
-	HeapTuple	tuple;
-
-	if (!pg_dbms_stats_use_locked_stats)
-		goto next_plugin;
-
-	tuple = get_merged_column_stats(indexOid, indexattnum, false);
-	vardata->statsTuple = tuple;
-	if (tuple == NULL)
-		goto next_plugin;
-
-	vardata->freefunc = FreeHeapTuple;
-			
-#ifdef PGDS_HAVE_ACL_OK
-	/*
-	 * set acl_ok if required. See the definition of set_acl_ok for details.
-	 */
-	if (set_acl_ok)
+	if (pg_dbms_stats_use_locked_stats)
 	{
-		/*
-		 * XXX: we had to scan the whole the rel array since we got
-		 * only the oid of the index.
-		 */
-		int i;
+		HeapTuple	tuple;
 
-		/* don't stop by this misassumption */
-		if (root->simple_rel_array == NULL)
+		tuple = get_merged_column_stats(indexOid, indexattnum, false);
+		if (tuple != NULL)
 		{
-			elog(WARNING, "pg_dbms_stats internal error. relation has not been set up. index %d ignored", indexOid);
-			goto next_plugin;
-		}
-
-		/*
-		 * scan over simple_rel_array_size to find the owner relation of the
-		 * index with the oid
-		 */
-		for (i = 1 ; i < root->simple_rel_array_size ; i++)
-		{
-			ListCell *lc;
-
-			foreach (lc, root->simple_rel_array[i]->indexlist)
-			{
-				IndexOptInfo *index = (IndexOptInfo *) lfirst(lc);
-				RangeTblEntry *rte;
-
-				if (index->indexoid != indexOid)
-					continue;
-
-				/* This relation is the owner of the given index, go ahead */
-				rte = planner_rt_fetch(index->rel->relid, root);
-
-				/* don't stop by this error */
-				if (rte->rtekind != RTE_RELATION)
-				{
-					elog(WARNING, "pg_dbms_stats internal error. index %d is owned by a non-relation", indexOid);
-					goto next_plugin;
-				}
-
-				vardata->acl_ok =
-					(pg_class_aclcheck(rte->relid, GetUserId(),
-									   ACL_SELECT) == ACLCHECK_OK);
-				break;
-			}
+			vardata->statsTuple = tuple;
+			vardata->freefunc = FreeHeapTuple;
+			return true;
 		}
 	}
-#endif
-	return true;
 
-next_plugin:
 	if (prev_get_index_stats)
 		return prev_get_index_stats(root, indexOid, indexattnum, vardata);
 	else
 		return false;
 }
 
-
-/*
- * dbms_stats_planner
- *   Hook function for planner_hook which cleans up invalidated statistics.
- */
-static PlannedStmt *
-dbms_stats_planner(Query *parse, int cursorOptions, ParamListInfo boundParams)
-{
-	PlannedStmt *ret;
-
-	cleanup_invalidated_cache();
-
-	if (prev_planner_hook)
-		ret = (*prev_planner_hook) (parse, cursorOptions, boundParams);
-	else
-		ret = standard_planner(parse, cursorOptions, boundParams);
-
-	cleanup_invalidated_cache();
-
-	return ret;
-}
-
-
 /*
  * Extract binary value from given column.
  */
@@ -1124,53 +819,38 @@ get_merged_relation_stats(Oid relid, BlockNumber *pages, double *tuples,
 	if (!found)
 		init_rel_stats_entry(entry, relid);
 
-	if (entry->valid)
-	{
-		/*
-		 * Valid entry with invalid relpage is a negative cache, which
-		 * eliminates excessive SPI calls below. Negative caches will be
-		 * invalidated again on invalidation of system relation cache, which
-		 * occur on modification of the dummy stats tables
-		 * dbms_stats.relation_stats_locked and column_stats_locked.
-		 */
-		if (entry->relpages == InvalidBlockNumber)
-			return;
-	}
+	/*
+	 * If we don't have valid cache entry, retrieve stats from catalogs,
+	 * pg_class and dbms_stats.relation_stats_locked, and merge them for planner use.
+	 * We also cache curpages value to make plans stable.
+	 */
 	if (!entry->valid)
 	{
-		/*
-		 * If we don't have valid cache entry, retrieve system stats and dummy
-		 * stats in dbms_stats.relation_stats_locked, then merge them for
-		 * planner use.  We also cache curpages value to make plans stable.
-		 */
 		bool		has_dummy;
 
 		PG_TRY();
 		{
 			++nested_level;
+
 			SPI_connect();
 
 			/*
-			 * Retrieve per-relation dummy statistics from
-			 * relation_stats_locked table via SPI.
+			 * Retrieve per-tuple dummy statistics from relation_stats_locked table
+			 * via SPI.  If we don't have dummy statistics, we return false to
+			 * tell caller thet we gave up to provide alternative statistics.
 			 */
 			has_dummy = execute_plan(&rows_plan, rows_query, relid, NULL, true);
-			if (!has_dummy)
+			if (has_dummy)
 			{
-				/* If dummy stats is not found, store negative cache. */
-				entry->relpages = InvalidBlockNumber;
-			}
-			else
-			{
-				/*
-				 * Retrieve per-relation system stats from pg_class.  We use
-				 * syscache to support indexes
-				 */
 				HeapTuple	tuple;
 				Form_pg_class form;
 				bool		isnull;
 				Datum		val;
 
+				/*
+				 * Retrieve per-relation stats from pg_class system catalog.
+				 * We use syscache to support indexes
+				 */
 				tuple = SearchSysCache1(RELOID, ObjectIdGetDatum(relid));
 				if (!HeapTupleIsValid(tuple))
 					elog(ERROR, "cache lookup failed for relation %u", relid);
@@ -1179,21 +859,27 @@ get_merged_relation_stats(Oid relid, BlockNumber *pages, double *tuples,
 				/* Choose dummy or authentic */
 				val = get_binary_datum(1, &isnull);
 				entry->relpages = isnull ? form->relpages :
-					(BlockNumber) DatumGetInt32(val);
+						(BlockNumber) DatumGetInt32(val);
 				val = get_binary_datum(2, &isnull);
 				entry->reltuples = isnull ? form->reltuples :
-					(double) DatumGetFloat4(val);
+						(double) DatumGetFloat4(val);
 				val = get_binary_datum(3, &isnull);
 				entry->curpages = isnull ? InvalidBlockNumber :
-					(BlockNumber) DatumGetInt32(val);
+						(BlockNumber) DatumGetInt32(val);
+#if PG_VERSION_NUM >= 90200
 				val = get_binary_datum(4, &isnull);
 				entry->relallvisible = isnull ? form->relallvisible :
-					(BlockNumber) DatumGetInt32(val);
+						(BlockNumber) DatumGetInt32(val);
+#endif
 
 				ReleaseSysCache(tuple);
+
+				/* Mark this entry as valid. */
+				entry->valid = true;
 			}
-			entry->valid = true;
+
 			SPI_finish();
+
 			--nested_level;
 		}
 		PG_CATCH();
@@ -1204,14 +890,13 @@ get_merged_relation_stats(Oid relid, BlockNumber *pages, double *tuples,
 		PG_END_TRY();
 
 		/*
-		 * If no dummy statistics available for this relation, do nothing then
-		 * return immediately.
+		 * If we don't dummy statistics, tell caller that we didn't change
+		 * stats in Relation, and it's ok to use authentic statistics as-is.
 		 */
 		if (!has_dummy)
 			return;
 	}
 
-	/* Tweaking statistics using merged statistics */
 	if (!estimate)
 	{
 		*pages = entry->relpages;
@@ -1220,13 +905,15 @@ get_merged_relation_stats(Oid relid, BlockNumber *pages, double *tuples,
 	}
 
 	/*
-	 * Get current number of pages to estimate current number of tuples, based
-	 * on tuple density at the last ANALYZE and current number of pages.
+	 * Get current # of pages to estimate current # of tuples, based on
+	 * tuple density at the last ANALYZE and current # of pages.
 	 */
 	rel = relation_open(relid, NoLock);
 	rel->rd_rel->relpages = entry->relpages;
 	rel->rd_rel->reltuples = entry->reltuples;
+#if PG_VERSION_NUM >= 90200
 	rel->rd_rel->relallvisible = entry->relallvisible;
+#endif
 	dbms_stats_estimate_rel_size(rel, NULL, pages, tuples, allvisfrac,
 								 entry->curpages);
 	relation_close(rel, NoLock);
@@ -1253,96 +940,65 @@ get_merged_avgwidth(Oid relid, AttrNumber attnum)
 
 /*
  * get_merged_column_stats
- *   returns per-column statistics for given column
+ *   get per-column statistic of given column
  *
- * This caches the result to avoid excessive SPI calls for repetitive
- * request for every columns many time.
+ * We cache the result because this function is called for every used column.
  */
 static HeapTuple
 get_merged_column_stats(Oid relid, AttrNumber attnum, bool inh)
 {
 	HeapTuple		tuple;
 	HeapTuple		statsTuple;
-	bool			negative = false;
 
 	if (nested_level > 0 || relid < FirstNormalObjectId)
 		return NULL;	/* avoid recursive call and system objects */
 
-	/*
-	 * Return NULL for system catalog, directing the caller to use system
-	 * statistics.
-	 */
+	/* If the relation is system catalog, return NULL to use true statistics. */
 	if (dbms_stats_is_system_catalog_internal(relid))
 		return NULL;
 
 	/* Return cached statistics, if any. */
-	if ((tuple = column_cache_search(relid, attnum, inh, &negative)) != NULL)
+	if ((tuple = column_cache_search(relid, attnum, inh)) != NULL)
 		return tuple;
 
-	/* Obtain system statistics from syscache. */
+	/* Obtain statistics from syscache.  This must be released afterward. */
 	statsTuple = SearchSysCache3(STATRELATTINH,
 								 ObjectIdGetDatum(relid),
 								 Int16GetDatum(attnum),
 								 BoolGetDatum(inh));
-	if (negative)
+	PG_TRY();
 	{
-		/*
-		 * Return system statistics whatever it is if negative cache for this
-		 * column is returned
-		 */
-		tuple = heap_copytuple(statsTuple);
+		++nested_level;
+
+		SPI_connect();
+
+		/* Obtain dummy statistics. */
+		if (execute_plan(&tuple_plan, tuple_query, relid, &attnum, inh))
+		{
+			/* merge the dummy statistics and the true statistics */
+			tuple = dbms_stats_merge_internal(SPI_tuptable->vals[0], statsTuple,
+											  SPI_tuptable->tupdesc);
+		}
+		else
+		{
+			/* if dummy statistics does not exist, use the true statistics */
+			tuple = statsTuple;
+		}
+
+		/* Cache merged result for subsequent calls. */
+		if (tuple != NULL)
+			tuple = column_cache_enter(tuple);
+
+		SPI_finish();
+
+		--nested_level;
 	}
-	else
+	PG_CATCH();
 	{
-		/*
-		 * Search for dummy statistics and try merge with system stats.
-		 */
-		PG_TRY();
-		{
-			/*
-			 * Save current context in order to use during SPI is
-			 * connected.
-			 */
-			MemoryContext outer_cxt = CurrentMemoryContext;
-			bool		  exec_success;
-
-			++nested_level;
-			SPI_connect();
-
-			/* Obtain dummy statistics for the column using SPI call. */
-			exec_success = 
-				execute_plan(&tuple_plan, tuple_query, relid, &attnum, inh);
-
-			/* Reset to the outer memory context for following steps. */
-			MemoryContextSwitchTo(outer_cxt);
-			
-			if (exec_success)
-			{
-				/* merge the dummy statistics with the system statistics */
-				tuple = dbms_stats_merge_internal(SPI_tuptable->vals[0],
-												  statsTuple,
-												  SPI_tuptable->tupdesc);
-			}
-			else
-				tuple = NULL;
-
-			/* Cache merged result for subsequent calls. */
-			tuple = column_cache_enter(relid, attnum, inh, tuple);
-
-			/* Return system stats if the merging results in failure. */
-			if (!HeapTupleIsValid(tuple))
-				tuple = heap_copytuple(statsTuple);
-
-			SPI_finish();
-			--nested_level;
-		}
-		PG_CATCH();
-		{
-			--nested_level;
-			PG_RE_THROW();
-		}
-		PG_END_TRY();
+		--nested_level;
+		PG_RE_THROW();
 	}
+	PG_END_TRY();
 
 	if (HeapTupleIsValid(statsTuple))
 		ReleaseSysCache(statsTuple);
@@ -1355,89 +1011,74 @@ get_merged_column_stats(Oid relid, AttrNumber attnum, bool inh)
  *   Search statistic of the given column from the cache.
  */
 static HeapTuple
-column_cache_search(Oid relid, AttrNumber attnum, bool inh, bool *negative)
+column_cache_search(Oid relid, AttrNumber attnum, bool inh)
 {
 	StatsRelationEntry *entry;
 	bool			found;
 	ListCell	   *lc;
 
-	*negative = false;
 	/*
 	 * First, get cached relation stats.  If we have not cached relation stats,
 	 * we don't have column stats too.
 	 */
 	entry = hash_search(rel_stats, &relid, HASH_FIND, &found);
-	if (!found)
+	if (!found )
 		return NULL;
 
 	/*
-	 * We assume that not so many column_stats_effective are defined on one
-	 * relation, so we use simple linear-search here.  Hash table would be an
-	 * alternative, but it seems overkill so far.
+	 * It is assumed that not so many column_stats_effective are defined on a relation, so we
+	 * use simple linear-search here.  Using hash table for each relation would
+	 * be another solution, but it seems overkill so far.
 	 */
 	foreach(lc, entry->col_stats)
 	{
-		StatsColumnEntry *ent = (StatsColumnEntry*) lfirst (lc);
+		HeapTuple	tuple = (HeapTuple) lfirst(lc);
+		Form_pg_statistic form = get_pg_statistic(tuple);
 
-		if (ent->attnum != attnum || ent->inh != inh) continue;
-
-		if (ent->negative)
-		{
-			/* Retrun NULL for negative cache, with noticing of that.*/
-			*negative = true;
-			return NULL;
-		}
-
-		return ent->tuple;
+		/* Find statistic of the given column from the cache. */
+		if (form->staattnum == attnum && form->stainherit == inh)
+			return tuple;
 	}
 
-	return NULL;	/* Not yet registered. */
+	return NULL;	/* not found */
 }
 
 /*
- * Cache a per-column statistics. Storing in CacheMemoryContext, the cached
- * statistics will live through the current session, unless dummy statistics or
- * table definition have been changed.
+ * Cache a per-column statistics. Cached statistics are valid through the
+ * current session, unless dummy statistics or table definition have been
+ * changed.
+ *
+ * We store statistics tuple in longer-lifetime context, CacheMemoryContext, to
+ * keep them alive during this session.
  */
 static HeapTuple
-column_cache_enter(Oid relid, int32 attnum, bool inh, HeapTuple tuple)
+column_cache_enter(HeapTuple tuple)
 {
 	MemoryContext	oldcontext;
-	StatsColumnEntry *newcolent;
+	HeapTuple		newtuple;
+	Form_pg_statistic form;
 	StatsRelationEntry *entry;
 	bool			found;
 
-	Assert(tuple == NULL || !heap_attisnull(tuple, 1));
+	Assert(tuple != NULL);
+	Assert(!heap_attisnull(tuple, 1));
 
-	entry = hash_search(rel_stats, &relid, HASH_ENTER, &found);
+	form = get_pg_statistic(tuple);
+
+	entry = hash_search(rel_stats, &form->starelid, HASH_ENTER, &found);
 	if (!found)
-		init_rel_stats_entry(entry, relid);
+		init_rel_stats_entry(entry, form->starelid);
 
 	/*
-	 * Adding this column stats to the column stats list of the relation stats
-	 * cache just obtained.
+	 * Link new column statistics entry into the list in relation statsstics
+	 * entry.
 	 */
 	oldcontext = MemoryContextSwitchTo(CacheMemoryContext);
-	newcolent = (StatsColumnEntry*)palloc(sizeof(StatsColumnEntry));
-	newcolent->attnum = attnum;
-	newcolent->inh = inh;
-
-	if (HeapTupleIsValid(tuple))
-	{
-		newcolent->negative = false;
-		newcolent->tuple = heap_copytuple(tuple);
-	}
-	else
-	{
-		/* Invalid tuple makes a negative cache. */
-		newcolent->negative = true;
-		newcolent->tuple = NULL;
-	}
-
-	entry->col_stats = lappend(entry->col_stats, newcolent);
+	newtuple = heap_copytuple(tuple);
+	entry->col_stats = lappend(entry->col_stats, newtuple);
 	MemoryContextSwitchTo(oldcontext);
 
-	return newcolent->tuple;
+	return newtuple;
 }
 
 /*
@@ -1458,144 +1099,62 @@ execute_plan(SPIPlanPtr *plan,
 	int		nargs;
 	Datum	values[3];
 	bool	nulls[3] = { false, false, false };
-	Oid			save_userid;
-	int			save_sec_context;
 
-	/* XXXX: this works for now but should be fixed later.. */
 	nargs = (attnum ? 3 : 1);
 
-	/*
-	 * The dummy statistics table allows access from no one other than its
-	 * owner or superuser.
-	 */
-	GetUserIdAndSecContext(&save_userid, &save_sec_context);
-	SetUserIdAndSecContext(get_stats_table_owner(),
-						   save_sec_context | SECURITY_LOCAL_USERID_CHANGE);
-
-	PG_TRY();
+	/* When plan is not given, create plan from query string at first. */
+	if (*plan == NULL)
 	{
-		/* Create plan from the query if not yet. */
-		if (*plan == NULL)
-		{
-			*plan = SPI_prepare(query, nargs, argtypes);
-			if (*plan == NULL)
-				elog(ERROR,
-					 "pg_dbms_stats: SPI_prepare() failed. result = %d",
-					 SPI_result);
-
-			SPI_keepplan(*plan);
-		}
-
-		values[0] = ObjectIdGetDatum(relid);
-		values[1] = Int16GetDatum(attnum ? *attnum : 0);
-		values[2] = BoolGetDatum(inh);
-
-		ret = SPI_execute_plan(*plan, values, nulls, true, 1);
+		SPIPlanPtr p;
+		p = SPI_prepare(query, nargs, argtypes);
+		if (p == NULL)
+			elog(ERROR, "pg_dbms_stats: SPI_prepare => %d", SPI_result);
+		*plan = SPI_saveplan(p);
+		SPI_freeplan(p);
 	}
-	PG_CATCH();
-	{
-		SetUserIdAndSecContext(save_userid, save_sec_context);
-		if (geterrcode() == ERRCODE_INSUFFICIENT_PRIVILEGE)
-			errdetail("dbms_stats could not access the object as the role \"%s\".",
-				stats_table_owner_name);
-		errhint("Check your settings of pg_dbms_stats.");
-		PG_RE_THROW();
-	}
-	PG_END_TRY();
 
-	SetUserIdAndSecContext(save_userid, save_sec_context);
+	values[0] = ObjectIdGetDatum(relid);
+	values[1] = Int16GetDatum(attnum ? *attnum : 0);
+	values[2] = BoolGetDatum(inh);
+
+	ret = SPI_execute_plan(*plan, values, nulls, true, 1);
 	if (ret != SPI_OK_SELECT)
-		elog(ERROR, "pg_dbms_stats: SPI_execute_plan() returned %d", ret);
+		elog(ERROR, "pg_dbms_stats: SPI_execute_plan => %d", ret);
 
 	return SPI_processed > 0;
 }
 
 /*
- * statscache_rel_callback
+ * StatsCacheRelCallback
  *		Relcache inval callback function
  *
- * Invalidates cached statistics of the given relid, or all cached statistics
- * if relid == InvalidOid. The statsTuple in the hash entries are directly
- * passed to planner so we cannot remove them until planner ends. Just mark
- * here then cleanup after planner finishes work.
- */
-static void
-statscache_rel_callback(Datum arg, Oid relid)
-{
-	StatsRelationEntry *entry;
-
-	if (relid != InvalidOid)
-	{
-		bool found;
-
-		/*
-		 * invalidate the entry for the specfied relation. Don't mind if found.
-		 */
-		entry = hash_search(rel_stats, &relid, HASH_FIND, &found);
-		if (found)
-		{
-			entry->invalidated = true;
-			rel_invalidated = true;
-		}
-	}
-	else
-	{
-		/* invalidate all the entries of the hash */
-		HASH_SEQ_STATUS		status;
-
-		hash_seq_init(&status, rel_stats);
-		while ((entry = hash_seq_search(&status)) != NULL)
-		{
-			entry->invalidated = true;
-			rel_invalidated = true;
-		}
-	}
-}
-
-/*
- * cleanup_invalidated_cache()
- *		Cleanup invalidated stats cache
+ * Invalidate cached statistic info of the given relid, or all cached statistic
+ * info if relid == InvalidOid.  We don't complain even when we don't have such
+ * statistics.
  *
- * removes invalidated cache entries.
+ * Note: arg is not used.
  */
 static void
-cleanup_invalidated_cache(void)
+StatsCacheRelCallback(Datum arg, Oid relid)
 {
 	HASH_SEQ_STATUS		status;
 	StatsRelationEntry *entry;
 
-	/* Return immediately if nothing to do */
-	if (!rel_invalidated)
-		return;
-
-	/*
-	 * Reset rel_invalidated first so that we don't lose invalidations that
-	 * happens during this round of cleanup.
-	 */
-	rel_invalidated = false;
-
 	hash_seq_init(&status, rel_stats);
 	while ((entry = hash_seq_search(&status)) != NULL)
 	{
-		ListCell	*lc;
-
-		if (!entry->invalidated)
-			continue;
-
-		/* Discard every column statistics */
-		foreach (lc, entry->col_stats)
+		if (relid == InvalidOid || relid == entry->relid)
 		{
-			StatsColumnEntry *ent = (StatsColumnEntry*) lfirst(lc);
+			/* Mark the relation entry as INVALID */
+			entry->valid = false;
 
-			if (!ent->negative)
-				pfree(ent->tuple);
-			pfree(ent);
+			/* Discard every column statistics */
+			list_free_deep(entry->col_stats);
+			entry->col_stats = NIL;
 		}
-		list_free(entry->col_stats);
-
-		/* Finally remove the hash entry. */
-		hash_search(rel_stats, &entry->relid, HASH_REMOVE, NULL);
 	}
+
+	/* We always check throughout the list, so hash_seq_term is not necessary */
 }
 
 /*
@@ -1618,7 +1177,7 @@ init_rel_stats(void)
 	ctl.hcxt = CacheMemoryContext;
 	hash = hash_create("dbms_stats relation statistics cache",
 					   MAX_REL_CACHE,
-					   &ctl, HASH_ELEM | HASH_CONTEXT | HASH_FUNCTION);
+					   &ctl, HASH_ELEM | HASH_CONTEXT);
 
 	rel_stats = hash;
 }
@@ -1632,7 +1191,6 @@ init_rel_stats_entry(StatsRelationEntry *entry, Oid relid)
 {
 	entry->relid = relid;
 	entry->valid = false;
-	entry->invalidated = false;
 	entry->relpages = InvalidBlockNumber;
 	entry->reltuples = 0.0;
 	entry->relallvisible = InvalidBlockNumber;
@@ -1652,12 +1210,12 @@ init_rel_stats_entry(StatsRelationEntry *entry, Oid relid)
  * the attribute widths for estimation purposes.
  *
  * Note: This function is copied from plancat.c in core source tree of version
- * 9.2, and customized for pg_dbms_stats.  Changes from original one are:
+ * 9.2, and customized for pg_dbms_stats.  Changes from orignal one are:
  *   - rename by prefixing dbms_stats_
  *   - add 3 parameters (relpages, reltuples, curpage) to pass dummy curpage
  *     values.
  *   - Get current # of pages only when supplied curpages is InvalidBlockNumber
- *   - get fraction of all-visible-pages
+ *   - get franction of all-visible-pages
  */
 static void
 dbms_stats_estimate_rel_size(Relation rel, int32 *attr_widths,
@@ -1673,9 +1231,6 @@ dbms_stats_estimate_rel_size(Relation rel, int32 *attr_widths,
 	{
 		case RELKIND_RELATION:
 		case RELKIND_INDEX:
-#if PG_VERSION_NUM >= 90300
-		case RELKIND_MATVIEW:
-#endif
 		case RELKIND_TOASTVALUE:
 			/* it has storage, ok to call the smgr */
 			if (curpages == InvalidBlockNumber)
@@ -1729,8 +1284,11 @@ dbms_stats_estimate_rel_size(Relation rel, int32 *attr_widths,
 			/* coerce values in pg_class to more desirable types */
 			relpages = (BlockNumber) rel->rd_rel->relpages;
 			reltuples = (double) rel->rd_rel->reltuples;
+#if PG_VERSION_NUM >= 90200
 			relallvisible = (BlockNumber) rel->rd_rel->relallvisible;
-
+#else
+			relallvisible = 0;
+#endif
 			/*
 			 * If it's an index, discount the metapage while estimating the
 			 * number of tuples.  This is a kluge because it assumes more than
@@ -1832,7 +1390,7 @@ dbms_stats_get_rel_data_width(Relation rel, int32 *attr_widths)
 
 	for (i = 1; i <= RelationGetNumberOfAttributes(rel); i++)
 	{
-		Form_pg_attribute att = get_attrs(rel->rd_att->attrs[i - 1]);
+		Form_pg_attribute att = rel->rd_att->attrs[i - 1];
 		int32		item_width;
 
 		if (att->attisdropped)
