@@ -13,7 +13,7 @@ SELECT relid::regclass,
        reltuples,
        relallvisible,
        curpages
-  FROM dbms_stats.relation_stats_locked
+  FROM dbms_stats._relation_stats_locked
  ORDER BY relid;
 CREATE VIEW relations_backup_v AS
 SELECT id,
@@ -41,7 +41,7 @@ SELECT starelid::regclass, staattnum, attname, stainherit,
        staop1, staop2, staop3, staop4, staop5,
        stanumbers1, stanumbers2, stanumbers3, stanumbers4, stanumbers5,
        stavalues1, stavalues2, stavalues3, stavalues4, stavalues5
-  FROM dbms_stats.column_stats_locked c
+  FROM dbms_stats._column_stats_locked c
   JOIN pg_attribute a
     ON (c.starelid = a.attrelid AND c.staattnum = a.attnum)
  ORDER BY starelid, staattnum, stainherit;
@@ -103,4 +103,3 @@ SELECT nspname, relname, relpages, reltuples, relallvisible,
        stavalues1, stavalues2, stavalues3, stavalues4, stavalues5
   FROM dbms_stats.work
  ORDER BY nspname, relname, attname, stainherit;
-ANALYZE s0.sft0;
